#!/usr/bin/env python

"""
A processing framework for iMIP content.

Copyright (C) 2014, 2015, 2016, 2017 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from email import message_from_file
from imiptools.config import settings
from imiptools.client import Client
from imiptools.content import handle_itip_part, have_itip_part, \
                              is_returned_message
from imiptools.data import get_address, get_addresses, get_uri
from imiptools.mail import Messenger
from imiptools.stores import get_store, get_publisher, get_journal
import sys, os

# Postfix exit codes.

EX_TEMPFAIL     = 75

# Processing of incoming messages.

def get_all_values(msg, key):

    "Return all values in 'msg' for 'key'."

    l = []
    for v in msg.get_all(key) or []:
        for s in v.split(","):
            l.append(s.strip())
    return l

class Processor:

    "The processing framework."

    def __init__(self, handlers, outgoing_only=False):
        self.handlers = handlers
        self.outgoing_only = outgoing_only
        self.messenger = None
        self.lmtp_socket = None
        self.store_type = None
        self.store_dir = None
        self.publishing_dir = None
        self.journal_dir = None
        self.preferences_dir = None
        self.debug = False

    def get_store(self):
        return get_store(self.store_type, self.store_dir)

    def get_publisher(self):
        return get_publisher(self.publishing_dir)

    def get_journal(self):
        return get_journal(self.store_type, self.journal_dir)

    def process(self, f, original_recipients):

        """
        Process content from the stream 'f' accompanied by the given
        'original_recipients'.
        """

        msg = message_from_file(f)

        messenger = self.messenger
        store = self.get_store()
        publisher = self.get_publisher()
        journal = self.get_journal()
        preferences_dir = self.preferences_dir

        # Handle messages with iTIP parts.
        # Typically, the details of recipients are of interest in handling
        # messages.

        if not self.outgoing_only:
            senders = get_addresses(get_all_values(msg, "Reply-To") or get_all_values(msg, "From") or [])
            original_recipients = original_recipients or get_addresses(get_all_values(msg, "To") or [])
            users = []

            for recipient in original_recipients:
                users.append(get_uri(recipient))

        # However, outgoing messages do not usually presume anything about the
        # eventual recipients and focus on the sender instead. If possible, the
        # sender is identified, but since this may be the calendar system (and
        # the actual sender is defined in the object), and since the recipient
        # may be in a Bcc header that is not available here, it may be left as
        # None and deduced from the object content later. 

        else:
            senders = []

            for sender in get_addresses(get_all_values(msg, "From") or []):
                if sender != settings["MESSAGE_SENDER"]:
                    senders.append(sender)

            users = [senders and get_uri(senders[0]) or None]

        # Process the message for each recipient.

        if self.debug:
            print >>sys.stderr, "Recipients..."
            print >>sys.stderr, ", ".join(map(str, users))

        for user in users:
            Recipient(user, messenger, store, publisher, journal,
                      preferences_dir, self.handlers, self.outgoing_only,
                      self.debug).process(msg, senders)

    def process_args(self, args, stream):

        """
        Interpret the given program arguments 'args' and process input from the
        given 'stream'.
        """

        args = parse_args(args, {"--show-config" : ("show_config", False)})

        self.debug = args["debug"]

        # Obtain arguments or configured defaults.

        self.store_type = args.get("store_type") or settings["STORE_TYPE"]
        self.store_dir = args.get("store_dir") or settings["STORE_DIR"]
        self.journal_dir = args.get("journal_dir") or settings["JOURNAL_DIR"]
        self.preferences_dir = args.get("preferences_dir") or settings["PREFERENCES_DIR"]
        self.publishing_dir = args.get("publishing_dir") or settings["PUBLISH_DIR"]

        self.messenger = Messenger(lmtp_socket=args["lmtp"],
                                   local_smtp=args["local_smtp"],
                                   sender=(args["senders"] or [None])[0])

        # Show configuration and exit if requested.

        if args["show_config"]:
            print """\
Store type: %s
Store directory: %s
Journal directory: %s
Preferences directory: %s
Publishing directory: %s""" % (
                self.store_type, self.store_dir, self.journal_dir,
                self.preferences_dir, self.publishing_dir)
            return

        # If debug mode is set, extend the line length for convenience.

        if self.debug:
            settings["CALENDAR_LINE_LENGTH"] = 1000
            print >>sys.stderr, "Store type", self.store_type, "at", self.store_dir

        # Process the input.

        self.process(stream, args["original_recipients"])

    def __call__(self):

        """
        Obtain arguments from the command line to initialise the processor
        before invoking it.
        """

        args = sys.argv[1:]

        # Show the help text if requested.

        if "--help" in args:
            show_help(os.path.split(sys.argv[0])[-1])

        # In debug mode, process the message without exception handling.

        elif "-d" in args:
            self.process_args(args, sys.stdin)

        # Otherwise, process the message and handle exceptions gracefully.

        else:
            try:
                self.process_args(args, sys.stdin)
            except SystemExit, value:
                sys.exit(value)
            except Exception, exc:
                if "-v" in args:
                    raise

                # Obtain the exception origin.

                type, value, tb = sys.exc_info()
                while tb.tb_next:
                    tb = tb.tb_next
                f = tb.tb_frame
                co = f and f.f_code
                filename = co and co.co_filename

                print >>sys.stderr, "Exception %s at %d in %s" % (exc, tb.tb_lineno, filename)

                #import traceback
                #traceback.print_exc(file=open("/tmp/mail.log", "a"))

                sys.exit(EX_TEMPFAIL)

        sys.exit(0)

class Recipient(Client):

    "A processor acting as a client on behalf of a recipient."

    def __init__(self, user, messenger, store, publisher, journal, preferences_dir,
                 handlers, outgoing_only, debug):

        """
        Initialise the recipient with the given 'user' identity, 'messenger',
        'store', 'publisher', 'journal', 'preferences_dir', 'handlers',
        'outgoing_only' and 'debug' status.
        """

        Client.__init__(self, user, messenger, store, publisher, journal, preferences_dir)
        self.handlers = handlers
        self.outgoing_only = outgoing_only
        self.debug = debug

    def process(self, msg, senders):

        """
        Process the given 'msg' for a single recipient, having the given
        'senders'.

        Processing individually means that contributions to resulting messages
        may be constructed according to individual preferences.
        """

        handlers = {}

        # Instantiate handlers for the supported methods.

        for name, cls in self.handlers:
            handlers[name] = cls(senders, self.user and get_address(self.user),
                                 self.messenger, self.store, self.publisher,
                                 self.journal, self.preferences_dir)

        handled = False

        # Check for participating recipients. Non-participating recipients will
        # have their messages left as being unhandled.

        if not is_returned_message(msg) and (self.outgoing_only or self.is_participating()):

            # Handle parts.

            for part in msg.walk():
                if self.debug and have_itip_part(part):
                    print >>sys.stderr, "Handle method %s..." % part.get_param("method")

                handled = handle_itip_part(part, handlers) or handled

        # When processing outgoing messages, no replies or deliveries are
        # performed.

        if self.outgoing_only:
            return

        # Get responses from the handlers.

        all_responses = []
        for handler in handlers.values():
            all_responses += handler.get_results()

        # Pack any returned parts into messages.

        if all_responses:
            outgoing_parts = {}
            forwarded_parts = []

            for outgoing_recipients, part in all_responses:
                if outgoing_recipients:
                    for outgoing_recipient in outgoing_recipients:
                        if not outgoing_parts.has_key(outgoing_recipient):
                            outgoing_parts[outgoing_recipient] = []
                        outgoing_parts[outgoing_recipient].append(part)
                else:
                    forwarded_parts.append(part)

            # Reply using any outgoing parts in a new message.

            if outgoing_parts:

                # Obtain free/busy details, if configured to do so.

                fb = self.can_provide_freebusy(handlers) and self.get_freebusy_part()

                for outgoing_recipient, parts in outgoing_parts.items():

                    # Bundle free/busy messages, if configured to do so.

                    if fb: parts.append(fb)
                    message = self.messenger.make_outgoing_message(parts, [outgoing_recipient])

                    if self.debug:
                        print >>sys.stderr, "Outgoing parts for %s..." % outgoing_recipient
                        print message
                    else:
                        self.messenger.sendmail([outgoing_recipient], message.as_string())

            # Forward messages to their recipients either wrapping the existing
            # message, accompanying it or replacing it.

            if forwarded_parts:

                # Determine whether to wrap, accompany or replace the message.

                prefs = self.get_preferences()
                incoming = prefs.get("incoming", settings["INCOMING_DEFAULT"])

                if incoming == "message-only":
                    messages = [msg]
                else:
                    summary = self.messenger.make_summary_message(msg, forwarded_parts)
                    if incoming == "summary-then-message":
                        messages = [summary, msg]
                    elif incoming == "message-then-summary":
                        messages = [msg, summary]
                    elif incoming == "summary-only":
                        messages = [summary]
                    else: # incoming == "summary-wraps-message":
                        messages = [self.messenger.wrap_message(msg, forwarded_parts)]

                for message in messages:
                    if self.debug:
                        print >>sys.stderr, "Forwarded parts..."
                        print message
                    elif self.messenger.local_delivery():
                        self.messenger.sendmail([get_address(self.user)], message.as_string())

        # Unhandled messages are delivered as they are.

        if not handled:
            if self.debug:
                print >>sys.stderr, "Unhandled parts..."
                print msg
            elif self.messenger.local_delivery():
                self.messenger.sendmail([get_address(self.user)], msg.as_string())

    def can_provide_freebusy(self, handlers):

        "Test for any free/busy information produced by 'handlers'."

        fbhandler = handlers.get("VFREEBUSY")
        if fbhandler:
            fbmethods = fbhandler.get_outgoing_methods()
            return not "REPLY" in fbmethods and not "PUBLISH" in fbmethods
        else:
            return False

# Standard arguments used by imip-agent programs.

def parse_args(args, extra_argdefs=None):

    """
    Interpret the given program arguments 'args'. Any 'extra_argdefs' define a
    mapping from option arguments to (option name, starting value) tuples to be
    considered in addition to (or as replacements for) the default definitions.
    """

    argdefs = {
        "-d" : ("debug", False),
        "-j" : ("journal_dir", None),
        "-l" : ("lmtp", None),
        "-L" : ("local_smtp", False),
        "-o" : ("original_recipients", []),
        "-p" : ("preferences_dir", None),
        "-P" : ("publishing_dir", None),
        "-s" : ("senders", []),
        "-S" : ("store_dir", None),
        "-T" : ("store_type", None),
        }

    argdefs.update(extra_argdefs)

    l = []
    option = None

    for arg in args:

        # Set any selected option value.

        if option and argdefs.has_key(option):
            name, value = argdefs[option]
            argdefs[option] = name, arg
            option = None

        # Where recognised, obtain the option name and value.

        elif argdefs.has_key(arg):
            name, value = argdefs[arg]

            # For boolean options, invert the current value.

            if isinstance(value, bool):
                argdefs[arg] = name, not value

            # For list options, switch to the given list and collect arguments.

            elif isinstance(value, list):
                l = value

            # Otherwise, select the option.

            else:
                option = arg

        # Where unrecognised, collect the argument in the current list.

        else:
            l.append(arg)

    # Return a mapping from option names to values.

    return dict(argdefs.values())

def show_help(progname):
    print >>sys.stderr, help_text % progname

help_text = """\
Usage: %s [ -o <recipient> ... ] [-s <sender> ... ] [ -l <socket> | -L ] \\
         [ -T <store type ] \\
         [ -S <store directory> ] [ -P <publishing directory> ] \\
         [ -p <preferences directory> ] [ -j <journal directory> ] \\
         [ -d ] [ --show-config ]

Address options:

-o  Indicate the original recipients of the message, overriding any found in
    the message headers
-s  Indicate the senders of the message, overriding any found in the message
    headers

Delivery options:

-l  The socket filename for LMTP communication with a mailbox solution,
    selecting the LMTP delivery method
-L  Selects the local SMTP delivery method, requiring a suitable mail system
    configuration

(Where a program needs to deliver messages, one of the above options must be
specified.)

Configuration options (overriding configured defaults):

-j  Indicates the location of quota-related journal information
-P  Indicates the location of published free/busy resources
-p  Indicates the location of user preference directories
-S  Indicates the location of the calendar data store containing user storage
    directories
-T  Indicates the store and journal type (the configured value if omitted)

Output options:

-d  Run in debug mode, producing informative output describing the behaviour
    of the program, displaying responses on standard output instead of sending
    messages

Diagnostic options:

--show-config  Show the configuration with the specified options and exit
               without performing any actions
"""

# vim: tabstop=4 expandtab shiftwidth=4
